"""
Implementation of the Chinese Remainder Theorem.

The code in this file is taken from SymPy (see the license below). We could
(and used to) just import sympy.ntheory.modular.crt, but SymPy is a very heavy
dependency for a rather basic algorithm.

Note: although many of the helper functions from SymPy are copied here, these
functions should not be used by code outside of ndindex. If you want to use
one of these algorithms, you should import them from SymPy (note that the
functions here have had their functionality stripped down from the SymPy
versions).

License
-------

Copyright (c) 2006-2021 SymPy Development Team

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of SymPy nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""

from math import gcd, prod

def gcdex(a, b):
    """Returns x, y, g such that g = x*a + y*b = gcd(a, b).

    Examples
    ========

    >>> from ndindex._crt import gcdex
    >>> gcdex(2, 3)
    (-1, 1, 1)
    >>> gcdex(10, 12)
    (-1, 1, 2)

    >>> x, y, g = gcdex(100, 2004)
    >>> x, y, g
    (-20, 1, 4)
    >>> x*100 + y*2004
    4

    """
    if (not a) and (not b):
        return (0, 1, 0)

    if not a:
        return (0, b//abs(b), abs(b))
    if not b:
        return (a//abs(a), 0, abs(a))

    if a < 0:
        a, x_sign = -a, -1
    else:
        x_sign = 1

    if b < 0:
        b, y_sign = -b, -1
    else:
        y_sign = 1

    x, y, r, s = 1, 0, 0, 1

    while b:
        (c, q) = (a % b, a // b)
        (a, b, r, s, x, y) = (b, c, x - q*r, y - q*s, r, s)

    return (x*x_sign, y*y_sign, a)

def _crt(U, M):
    """
    Chinese Remainder Theorem.

    Given a set of integer residues ``u_0,...,u_n`` and a set of
    co-prime integer moduli ``m_0,...,m_n``, returns an integer
    ``u``, such that ``u = u_i mod m_i`` for ``i = ``0,...,n``.

    Examples
    ========

    Consider a set of residues ``U = [49, 76, 65]``
    and a set of moduli ``M = [99, 97, 95]``. Then we have::

       >>> from ndindex._crt import _crt

       >>> _crt([49, 76, 65], [99, 97, 95])
       639985

    This is the correct result because::

       >>> [639985 % m for m in [99, 97, 95]]
       [49, 76, 65]

    Note: this is a low-level routine with no error checking.
    """
    p = prod(M)
    v = 0

    for u, m in zip(U, M):
        e = p // m
        s, _, _ = gcdex(e, m)
        v += e*(u*s % m)

    return v % p


def solve_congruence(*remainder_modulus_pairs):
    """Compute the integer ``n`` that has the residual ``ai`` when it is
    divided by ``mi`` where the ``ai`` and ``mi`` are given as pairs to
    this function: ((a1, m1), (a2, m2), ...). If there is no solution,
    return None. Otherwise return ``n`` and its modulus.

    The ``mi`` values need not be co-prime.

    Examples
    ========

    >>> from ndindex._crt import solve_congruence

    What number is 2 mod 3, 3 mod 5 and 2 mod 7?

    >>> solve_congruence((2, 3), (3, 5), (2, 7))
    23
    >>> [23 % m for m in [3, 5, 7]]
    [2, 3, 2]

    If you prefer to work with all remainder in one list and
    all moduli in another, send the arguments like this:

    >>> solve_congruence(*zip((2, 3, 2), (3, 5, 7)))
    23

    The moduli need not be co-prime; in this case there may or
    may not be a solution:

    >>> solve_congruence((2, 3), (4, 6)) is None
    True

    >>> solve_congruence((2, 3), (5, 6))
    5

    """
    def combine(c1, c2):
        """Return the tuple (a, m) which satisfies the requirement
        that n = a + i*m satisfy n = a1 + j*m1 and n = a2 = k*m2.

        References
        ==========

        .. [1] https://en.wikipedia.org/wiki/Method_of_successive_substitution
        """
        a1, m1 = c1
        a2, m2 = c2
        a, b, c = m1, a2 - a1, m2
        g = gcd(a, b, c)
        a, b, c = [i//g for i in [a, b, c]]
        if a != 1:
            inv_a, _, g = gcdex(a, c)
            if g != 1:
                return None
            b *= inv_a
        a, m = a1 + m1*b, m1*c
        return a, m

    rm = remainder_modulus_pairs

    rv = (0, 1)
    for rmi in rm:
        rv = combine(rv, rmi)
        if rv is None:
            break
        n, m = rv
        n = n % m
    else:
        return n

def crt(m, v, check=True):
    r"""Chinese Remainder Theorem.

    The moduli in m are assumed to be pairwise coprime.  The output
    is then an integer f, such that f = v_i mod m_i for each pair out
    of v and m.

    If the moduli are not co-prime the correct result will be returned
    if/when the test of the result is found to be incorrect. This result
    will be None if there is no solution.

    The keyword ``check`` can be set to False if it is known that the moduli
    are coprime.

    Examples
    ========

    As an example consider a set of residues ``U = [49, 76, 65]``
    and a set of moduli ``M = [99, 97, 95]``. Then we have::

       >>> from ndindex._crt import crt

       >>> crt([99, 97, 95], [49, 76, 65])
       639985

    This is the correct result because::

       >>> [639985 % m for m in [99, 97, 95]]
       [49, 76, 65]

    If the moduli are not co-prime, you may receive an incorrect result
    if you use ``check=False``:

       >>> crt([12, 6, 17], [3, 4, 2], check=False)
       954
       >>> [954 % m for m in [12, 6, 17]]
       [6, 0, 2]
       >>> crt([12, 6, 17], [3, 4, 2]) is None
       True
       >>> crt([3, 6], [2, 5])
       5

    Note: the order of gf_crt's arguments is reversed relative to crt,
    and that solve_congruence takes residue, modulus pairs.

    Programmer's note: rather than checking that all pairs of moduli share
    no GCD (an O(n**2) test) and rather than factoring all moduli and seeing
    that there is no factor in common, a check that the result gives the
    indicated residuals is performed -- an O(n) operation.
    """
    result = _crt(v, m)

    if check:
        if not all(v % m == result % m for v, m in zip(v, m)):
            result = solve_congruence(*list(zip(v, m)))

    return result

def ilcm(a, b):
    """Computes integer least common multiple.

    Examples
    ========

    >>> from ndindex._crt import ilcm
    >>> ilcm(5, 10)
    10
    >>> ilcm(7, 3)
    21

    """
    if 0 in [a, b]:
        return 0
    return a // gcd(a, b) * b # since gcd(a,b) | a



import numbers
import itertools
from collections.abc import Sequence
from ._crt import prod

from .ndindex import ndindex, operator_index

class BroadcastError(ValueError):
    """
    Exception raised by :func:`iter_indices()` and
    :func:`broadcast_shapes()` when the input shapes are not broadcast
    compatible.

    """
    def __init__(self, arg1, shape1, arg2, shape2):
        self.arg1 = arg1
        self.shape1 = shape1
        self.arg2 = arg2
        self.shape2 = shape2

    def __str__(self):
        arg1, shape1, arg2, shape2 = self.args
        return f"shape mismatch: objects cannot be broadcast to a single shape.  Mismatch is between arg {arg1} with shape {shape1} and arg {arg2} with shape {shape2}."

class AxisError(ValueError, IndexError):
    """
    Exception raised by :func:`iter_indices()` and
    :func:`broadcast_shapes()` when the `skip_axes` argument is out of bounds.

    This is used instead of the NumPy exception of the same name so that
    `iter_indices` does not need to depend on NumPy.

    """
    def __init__(self, axis, ndim):
        # NumPy allows axis=-1 for 0-d arrays
        if (ndim < 0 or -ndim <= axis < ndim) and not (ndim == 0 and axis == -1):
            raise ValueError(f"Invalid AxisError ({axis}, {ndim})")
        self.axis = axis
        self.ndim = ndim

    def __str__(self):
        return f"axis {self.axis} is out of bounds for array of dimension {self.ndim}"

def broadcast_shapes(*shapes, skip_axes=()):
    """
    Broadcast the input shapes `shapes` to a single shape.

    This is the same as :external+numpy:py:func:`np.broadcast_shapes()
    <numpy.broadcast_shapes>`, except is also supports skipping axes in the
    shape with `skip_axes`.

    `skip_axes` can be a tuple of integers which apply to all shapes, or a
    list of tuples of integers, one for each shape, which apply to each
    respective shape. The `skip_axes` argument works the same as in
    :func:`iter_indices`. See its docstring for more details.

    If the shapes are not broadcast compatible (excluding `skip_axes`),
    :class:`BroadcastError` is raised.

    >>> from ndindex import broadcast_shapes
    >>> broadcast_shapes((2, 3), (3,), (4, 2, 1))
    (4, 2, 3)
    >>> broadcast_shapes((2, 3), (5,), (4, 2, 1))
    Traceback (most recent call last):
    ...
    ndindex.shapetools.BroadcastError: shape mismatch: objects cannot be broadcast to a single shape.  Mismatch is between arg 0 with shape (2, 3) and arg 1 with shape (5,).

    Axes in `skip_axes` apply to each shape *before* being broadcasted. Each
    shape will be broadcasted together with these axes removed. The dimensions
    in `skip_axes` do not need to be equal or broadcast compatible with one
    another. The final broadcasted shape be the result of broadcasting all the
    non-skip axes.

    >>> broadcast_shapes((10, 3, 2), (2, 20), skip_axes=[(0,), (1,)])
    (3, 2)

    """
    shapes = [asshape(shape, allow_int=False) for shape in shapes]
    skip_axes = normalize_skip_axes(shapes, skip_axes)

    if not shapes:
        return ()

    non_skip_shapes = [remove_indices(shape, skip_axis) for shape, skip_axis in zip(shapes, skip_axes)]
    dims = [len(shape) for shape in non_skip_shapes]
    N = max(dims)

    broadcasted_shape = [1]*N

    arg = None
    for i in range(-1, -N-1, -1):
        for j in range(len(shapes)):
            if dims[j] < -i:
                continue
            shape = non_skip_shapes[j]
            broadcasted_side = broadcasted_shape[i]
            shape_side = shape[i]
            if shape_side == 1:
                continue
            elif broadcasted_side == 1:
                broadcasted_side = shape_side
                arg = j
            elif shape_side != broadcasted_side:
                raise BroadcastError(arg, shapes[arg], j, shapes[j])
            broadcasted_shape[i] = broadcasted_side

    return tuple(broadcasted_shape)

def iter_indices(*shapes, skip_axes=(), _debug=False):
    """
    Iterate indices for every element of an arrays of shape `shapes`.

    Each shape in `shapes` should be a shape tuple, which are broadcast
    compatible along the non-skipped axes. Each iteration step will produce a
    tuple of indices, one for each shape, which would correspond to the same
    elements if the arrays of the given shapes were first broadcast together.

    This is a generalization of the NumPy :external+numpy:py:class:`np.ndindex()
    <numpy.ndindex>` function (which otherwise has no relation).
    `np.ndindex()` only iterates indices for a single shape, whereas
    `iter_indices()` supports generating indices for multiple broadcast
    compatible shapes at once. This is equivalent to first broadcasting the
    arrays then generating indices for the single broadcasted shape.

    Additionally, this function supports the ability to skip axes of the
    shapes using `skip_axes`. These axes will be fully sliced in each index.
    The remaining axes will be indexed one element at a time with integer
    indices.

    `skip_axes` should be a tuple of axes to skip or a list of tuples of axes
    to skip. If it is a single tuple, it applies to all shapes. Otherwise,
    each tuple applies to each shape respectively. It can use negative
    integers, e.g., `skip_axes=(-1,)` will skip the last axis. The order of
    the axes in `skip_axes` does not matter. Mixing negative and nonnegative
    skip axes is supported, but the skip axes must refer to unique dimensions
    for each shape.

    The axes in `skip_axes` refer to the shapes *before* broadcasting (if you
    want to refer to the axes after broadcasting, either broadcast the shapes
    and arrays first, or refer to the axes using negative integers). For
    example, `iter_indices((10, 2), (20, 1, 2), skip_axes=(0,))` will skip the
    size `10` axis of `(10, 2)` and the size `20` axis of `(20, 1, 2)`. The
    result is two sets of indices, one for each element of the non-skipped
    dimensions:

    >>> from ndindex import iter_indices
    >>> for idx1, idx2 in iter_indices((10, 2), (20, 1, 2), skip_axes=(0,)):
    ...     print(idx1, idx2)
    Tuple(slice(None, None, None), 0) Tuple(slice(None, None, None), 0, 0)
    Tuple(slice(None, None, None), 1) Tuple(slice(None, None, None), 0, 1)

    The skipped axes do not themselves need to be broadcast compatible, but
    the shapes with all the skipped axes removed should be broadcast
    compatible.

    For example, suppose `a` is an array with shape `(3, 2, 4, 4)`, which we
    wish to think of as a `(3, 2)` stack of 4 x 4 matrices. We can generate an
    iterator for each matrix in the "stack" with `iter_indices((3, 2, 4, 4),
    skip_axes=(-1, -2))`:

    >>> for idx in iter_indices((3, 2, 4, 4), skip_axes=(-1, -2)):
    ...     print(idx)
    (Tuple(0, 0, slice(None, None, None), slice(None, None, None)),)
    (Tuple(0, 1, slice(None, None, None), slice(None, None, None)),)
    (Tuple(1, 0, slice(None, None, None), slice(None, None, None)),)
    (Tuple(1, 1, slice(None, None, None), slice(None, None, None)),)
    (Tuple(2, 0, slice(None, None, None), slice(None, None, None)),)
    (Tuple(2, 1, slice(None, None, None), slice(None, None, None)),)

    .. note::

       The iterates of `iter_indices` are always a tuple, even if only a
       single shape is provided (one could instead use `for idx, in
       iter_indices(...)` above).

    As another example, say `a` is shape `(1, 3)` and `b` is shape `(2, 1)`,
    and we want to generate indices for every value of the broadcasted
    operation `a + b`. We can do this by using `a[idx1.raw] + b[idx2.raw]` for every
    `idx1` and `idx2` as below:

    >>> import numpy as np
    >>> a = np.arange(3).reshape((1, 3))
    >>> b = np.arange(100, 111, 10).reshape((2, 1))
    >>> a
    array([[0, 1, 2]])
    >>> b
    array([[100],
           [110]])
    >>> for idx1, idx2 in iter_indices((1, 3), (2, 1)):
    ...     print(f"{idx1 = }; {idx2 = }; {(a[idx1.raw], b[idx2.raw]) = }") # doctest: +SKIPNP1
    idx1 = Tuple(0, 0); idx2 = Tuple(0, 0); (a[idx1.raw], b[idx2.raw]) = (np.int64(0), np.int64(100))
    idx1 = Tuple(0, 1); idx2 = Tuple(0, 0); (a[idx1.raw], b[idx2.raw]) = (np.int64(1), np.int64(100))
    idx1 = Tuple(0, 2); idx2 = Tuple(0, 0); (a[idx1.raw], b[idx2.raw]) = (np.int64(2), np.int64(100))
    idx1 = Tuple(0, 0); idx2 = Tuple(1, 0); (a[idx1.raw], b[idx2.raw]) = (np.int64(0), np.int64(110))
    idx1 = Tuple(0, 1); idx2 = Tuple(1, 0); (a[idx1.raw], b[idx2.raw]) = (np.int64(1), np.int64(110))
    idx1 = Tuple(0, 2); idx2 = Tuple(1, 0); (a[idx1.raw], b[idx2.raw]) = (np.int64(2), np.int64(110))
    >>> a + b
    array([[100, 101, 102],
           [110, 111, 112]])

    To include an index into the final broadcasted array, you can simply
    include the final broadcasted shape as one of the shapes (the function
    :func:`broadcast_shapes` is useful here).

    >>> np.broadcast_shapes((1, 3), (2, 1))
    (2, 3)
    >>> for idx1, idx2, broadcasted_idx in iter_indices((1, 3), (2, 1), (2, 3)):
    ...     print(broadcasted_idx)
    Tuple(0, 0)
    Tuple(0, 1)
    Tuple(0, 2)
    Tuple(1, 0)
    Tuple(1, 1)
    Tuple(1, 2)

    """
    skip_axes = normalize_skip_axes(shapes, skip_axes)
    shapes = [asshape(shape, allow_int=False) for shape in shapes]

    if not shapes:
        yield ()
        return

    shapes = [asshape(shape) for shape in shapes]
    S = len(shapes)

    iters = [[] for i in range(S)]
    broadcasted_shape = broadcast_shapes(*shapes, skip_axes=skip_axes)

    idxes = [-1]*S

    while any(i is not None for i in idxes):
        for s, it, shape, sk in zip(range(S), iters, shapes, skip_axes):
            i = idxes[s]
            if i is None:
                continue
            if -i > len(shape):
                if not shape:
                    pass
                elif len(shape) == len(sk):
                    # The whole shape is skipped. Just repeat the most recent slice
                    it[0] = ncycles(it[0], prod(broadcasted_shape))
                else:
                    # Find the first non-skipped axis and repeat by however
                    # many implicit axes are left in the broadcasted shape
                    for j in range(-len(shape), 0):
                        if j not in sk:
                            break
                    it[j] = ncycles(it[j], prod(broadcasted_shape[:len(sk)-len(shape)+len(broadcasted_shape)]))

                idxes[s] = None
                continue

            val = associated_axis(broadcasted_shape, i, sk)
            if i in sk:
                it.insert(0, [slice(None)])
            else:
                if val == 0:
                    return
                elif val != 1 and shape[i] == 1:
                    it.insert(0, ncycles(range(shape[i]), val))
                else:
                    it.insert(0, range(shape[i]))
            idxes[s] -= 1

    if _debug: # pragma: no cover
        print(f"{iters = }")
    for idxes in itertools.zip_longest(*[itertools.product(*i) for i in
                                         iters], fillvalue=()):
        yield tuple(ndindex(idx) for idx in idxes)

#### Internal helpers


def asshape(shape, axis=None, *, allow_int=True, allow_negative=False):
    """
    Cast `shape` as a valid NumPy shape.

    The input can be an integer `n` (if `allow_int=True`), which is equivalent
    to `(n,)`, or a tuple of integers.

    If the `axis` argument is provided, an `IndexError` is raised if it is out
    of bounds for the shape.

    The resulting shape is always a tuple of nonnegative integers. If
    `allow_negative=True`, negative integers are also allowed.

    All ndindex functions that take a shape input should use::

        shape = asshape(shape)

    or::

        shape = asshape(shape, axis=axis)

    """
    from .integer import Integer
    from .tuple import Tuple
    if isinstance(shape, (Tuple, Integer)):
        raise TypeError("ndindex types are not meant to be used as a shape - "
                        "did you mean to use the built-in tuple type?")

    if isinstance(shape, numbers.Number):
        if allow_int:
            shape = (operator_index(shape),)
        else:
            raise TypeError(f"expected sequence of integers, not {type(shape).__name__}")

    if not isinstance(shape, Sequence) or isinstance(shape, str):
        raise TypeError("expected sequence of integers" + allow_int*" or a single integer" + ", not " + type(shape).__name__)
    l = len(shape)

    newshape = []
    # numpy uses __getitem__ rather than __iter__ to index into shape, so we
    # match that
    for i in range(l):
        # Raise TypeError if invalid
        val = shape[i]
        if val is None:
            raise ValueError("unknonwn (None) dimensions are not supported")

        newshape.append(operator_index(shape[i]))

        if not allow_negative and val < 0:
            raise ValueError("unknown (negative) dimensions are not supported")

    if axis is not None:
        if len(newshape) <= axis:
            raise IndexError(f"too many indices for array: array is {len(shape)}-dimensional, but {axis + 1} were indexed")

    return tuple(newshape)

def associated_axis(broadcasted_shape, i, skip_axes):
    """
    Return the associated element of `broadcasted_shape` corresponding to
    `shape[i]` given `skip_axes`. If there is not such element (i.e., it's out
    of bounds), returns None.

    This function makes implicit assumptions about its input and is only
    designed for internal use.

    """
    skip_axes = sorted(skip_axes, reverse=True)
    if i >= 0:
        raise NotImplementedError
    if i in skip_axes:
        return None
    # We assume skip_axes are all negative and sorted
    j = i
    for sk in skip_axes:
        if sk >= i:
            j += 1
        else:
            break
    if ndindex(j).isvalid(len(broadcasted_shape)):
        return broadcasted_shape[j]
    return None

def remove_indices(x, idxes):
    """
    Return `x` with the indices `idxes` removed.

    This function is only intended for internal usage.
    """
    if isinstance(idxes, int):
        idxes = (idxes,)
    dim = len(x)
    _idxes = sorted({i if i >= 0 else i + dim for i in idxes})
    _idxes = [i - a for i, a in zip(_idxes, range(len(_idxes)))]
    _x = list(x)
    for i in _idxes:
        _x.pop(i)
    return tuple(_x)

def unremove_indices(x, idxes, *, val=None):
    """
    Insert `val` in `x` so that it appears at `idxes`.

    Note that idxes must be either all negative or all nonnegative.

    This function is only intended for internal usage.
    """
    if any(i >= 0 for i in idxes) and any(i < 0 for i in idxes):
        # A mix of positive and negative indices presents a fundamental
        # problem: sometimes the result is not unique. For example, x = [0];
        # idxes = [1, -1] could be satisfied by both [0, None] or [0, None,
        # None], depending on whether each index refers to a separate None or
        # not (note that both cases are supported by remove_indices(), because
        # there it is unambiguous). But even worse, in some cases, there may
        # be no way to satisfy the given requirement. For example, given x =
        # [0, 1, 2, 3]; idxes = [3, -3], there is no way to insert None into x
        # so that remove_indices(res, idxes) == x. To see this, simply observe
        # that there is no size list x such that remove_indices(x, [3, -3])
        # returns a tuple of size 4:
        #
        # >>> [len(remove_indices(list(range(n)), [3, -3])) for n in range(4, 10)]
        # [2, 3, 5, 5, 6, 7]
        raise NotImplementedError("Mixing both negative and nonnegative idxes is not yet supported")
    x = list(x)
    n = len(idxes) + len(x)
    _idxes = sorted({i if i >= 0 else i + n for i in idxes})
    for i in _idxes:
        x.insert(i, val)
    return tuple(x)

class ncycles:
    """
    Iterate `iterable` repeated `n` times.

    This is based on a recipe from the `Python itertools docs
    <https://docs.python.org/3/library/itertools.html#itertools-recipes>`_,
    but improved to give a repr, and to denest when it can. This makes
    debugging :func:`~.iter_indices` easier.

    This is only intended for internal usage.

    >>> from ndindex.shapetools import ncycles
    >>> ncycles(range(3), 2)
    ncycles(range(0, 3), 2)
    >>> list(_)
    [0, 1, 2, 0, 1, 2]
    >>> ncycles(ncycles(range(3), 3), 2)
    ncycles(range(0, 3), 6)

    """
    def __new__(cls, iterable, n):
        if n == 1:
            return iterable
        return object.__new__(cls)

    def __init__(self, iterable, n):
        if isinstance(iterable, ncycles):
            self.iterable = iterable.iterable
            self.n = iterable.n*n
        else:
            self.iterable = iterable
            self.n = n

    def __repr__(self):
        return f"ncycles({self.iterable!r}, {self.n!r})"

    def __iter__(self):
        return itertools.chain.from_iterable(itertools.repeat(tuple(self.iterable), self.n))

def normalize_skip_axes(shapes, skip_axes):
    """
    Return a canonical form of `skip_axes` corresponding to `shapes`.

    A canonical form of `skip_axes` is a list of tuples of integers, one for
    each shape in `shapes`, which are a unique set of axes for each
    corresponding shape.

    If `skip_axes` is an integer, this is basically `[(skip_axes,) for s
    in shapes]`. If `skip_axes` is a tuple, it is like `[skip_axes for s in
    shapes]`.

    The `skip_axes` must always refer to unique axes in each shape.

    The returned `skip_axes` will always be negative integers and will be
    sorted.

    This function is only intended for internal usage.

    """
    # Note: we assume asshape has already been called on the shapes in shapes
    if isinstance(skip_axes, Sequence):
        if skip_axes and all(isinstance(i, Sequence) for i in skip_axes):
            if len(skip_axes) != len(shapes):
                raise ValueError(f"Expected {len(shapes)} skip_axes")
            return [normalize_skip_axes([shape], skip_axis)[0] for shape, skip_axis in zip(shapes, skip_axes)]
        else:
            try:
                [operator_index(i) for i in skip_axes]
            except TypeError:
                raise TypeError("skip_axes must be an integer, a tuple of integers, or a list of tuples of integers")

    skip_axes = asshape(skip_axes, allow_negative=True)

    # From here, skip_axes is a single tuple of integers

    if not shapes and skip_axes:
        raise ValueError("skip_axes must be empty if there are no shapes")

    new_skip_axes = []
    err = None
    for shape in shapes:
        s = tuple(sorted(ndindex(i).reduce(len(shape), negative_int=True, axiserror=True).raw for i in skip_axes))
        if len(s) != len(set(s)):
            err = ValueError(f"skip_axes {skip_axes} are not unique for shape {shape}")
            # For testing
            err.skip_axes = skip_axes
            err.shape = shape
        new_skip_axes.append(s)
    if err:
        raise err
    return new_skip_axes



from collections.abc import Sequence
from itertools import chain, product

from .ndindex import ImmutableObject, operator_index, ndindex
from .tuple import Tuple
from .slice import Slice
from .integer import Integer
from .integerarray import IntegerArray
from .newaxis import Newaxis
from .shapetools import asshape
from .subindex_helpers import ceiling
from ._crt import prod

class ChunkSize(ImmutableObject, Sequence):
    """
    Represents a chunk size tuple.

    A chunk size is a tuple of length n where each element is either a
    positive integer or `None`. It represents a chunking of an array with n
    dimensions, where each corresponding dimension is chunked by the
    corresponding chunk size, or not chunked for `None` (note, `None` chunks
    are currently not yet implemented).

    For example, given a 3 dimensional chunk size of `(20, 20, None)` and an
    array of shape `(40, 30, 10)`, the array would be split into four chunks,
    corresponding to the indices `0:20,0:20,:`, `0:20,20:30,:`,
    `20:40,0:20,:`, and `20:40,20:30,:`. Note that the size of a chunk may be
    less than the total chunk size if the array shape is not a multiple of the
    chunk size in a given dimension.

    `ChunkSize` behaves like a `tuple`. For example, `chunk_size[0]` gives the
    first chunk dimension, and `len(chunk_size)` gives the number of
    dimensions of a chunk. Also, the input to ChunkSize should be a tuple,
    just as with the `tuple` constructor, even for single dimensional chunk
    sizes.

    >>> from ndindex import ChunkSize
    >>> ChunkSize((20, 30, 40))
    ChunkSize((20, 30, 40))
    >>> ChunkSize((2**12,))
    ChunkSize((4096,))

    """
    def _typecheck(self, chunk_size):
        # TODO: Also accept ChunkSize(1, 2, 3)?
        if isinstance(chunk_size, Tuple):
            raise TypeError("Tuple is not a valid input to ChunkSize. Use tuple instead.")
        args = []
        for i in chunk_size:
            if i is None:
                raise NotImplementedError("None in chunks is not supported yet")
                # args.append(i)
            else:
                try:
                    i = operator_index(i)
                except TypeError:
                    raise TypeError("Chunks must be positive integers or None")
                if i <= 0:
                    raise ValueError("Chunks must be positive integers")
                args.append(i)
        return (tuple(args),)

    def __hash__(self):
        return hash(self.args[0])

    # Methods for collections.abc.Sequence to make ChunkSize act like a tuple
    def __getitem__(self, *args):
        return self.args[0].__getitem__(*args)

    def __len__(self):
        return len(self.args[0])

    def num_chunks(self, shape):
        """
        Give the number of chunks for the given `shape`.

        This is the same as `len(list(self.indices(shape)))`, but much faster.
        `shape` must have the same number of dimensions as `self`.

        >>> from ndindex import ChunkSize
        >>> chunk_size = ChunkSize((10, 10, 10))
        >>> shape = (10000, 10000, 10000)
        >>> # len(list(chunk_size.indices(shape))) would be very slow, as
        >>> # would have to iterate all 1 billion chunks
        >>> chunk_size.num_chunks(shape)
        1000000000

        See Also
        ========

        ChunkSize.indices

        """
        shape = asshape(shape)
        d = [ceiling(i, c) for i, c in zip(shape, self)]
        if 0 in d:
            return 1
        return prod(d)

    def indices(self, shape):
        """
        Yield a set of ndindex indices for the chunks on an array of shape `shape`.

        `shape` should have the same number of dimensions as `self`. If the
        shape is not a multiple of the chunk size, some chunks will be
        truncated, so that :any:`len(idx.args[i]) <ndindex.Slice.__len__>` can
        be used to get the size of an indexed axis.

        For example, if `a` has shape `(10, 19)` and is chunked into chunks
        of shape `(5, 5)`:

        >>> from ndindex import ChunkSize
        >>> chunk_size = ChunkSize((5, 5))
        >>> for idx in chunk_size.indices((10, 19)):
        ...     print(idx)
        Tuple(slice(0, 5, 1), slice(0, 5, 1))
        Tuple(slice(0, 5, 1), slice(5, 10, 1))
        Tuple(slice(0, 5, 1), slice(10, 15, 1))
        Tuple(slice(0, 5, 1), slice(15, 19, 1))
        Tuple(slice(5, 10, 1), slice(0, 5, 1))
        Tuple(slice(5, 10, 1), slice(5, 10, 1))
        Tuple(slice(5, 10, 1), slice(10, 15, 1))
        Tuple(slice(5, 10, 1), slice(15, 19, 1))

        See Also
        ========

        ChunkSize.num_chunks

        """
        shape = asshape(shape)

        if len(shape) != len(self):
            raise ValueError("chunks dimensions must equal the array dimensions")
        d = [ceiling(i, c) for i, c in zip(shape, self)]
        if 0 in d:
            yield Tuple(*[Slice(0, bool(i)*chunk_size, 1) for i, chunk_size in zip(d, self)]).expand(shape)
        for p in product(*[range(i) for i in d]):
            # p = (0, 0, 0), (0, 0, 1), ...
            yield Tuple(*[Slice(chunk_size*i, min(chunk_size*(i + 1), n), 1)
                          for n, chunk_size, i in zip(shape, self, p)])

    def as_subchunks(self, idx, shape, *, _force_slow=None):
        """
        Split an index `idx` on an array of shape `shape` into subchunk indices.

        Yields indices `c`, where `c` is an index for the chunk that
        should be sliced. Only those `c` for which `idx` includes at least one
        element are yielded.

        That is to say, for each `c` index yielded, `a[c][idx.as_subindex(c)]`
        will give those elements of `a[idx]` that are part of the `c` chunk,
        and together they give all the elements of `a[idx]`. See also the
        docstring of :meth:`~ndindex.NDIndex.as_subindex`.

        This method is roughly equivalent to

        .. code:: python

           def as_subchunks(self, idx, shape):
               for c in self.indices(shape):
                   try:
                       index = idx.as_subindex(c)
                   except ValueError:
                       # as_subindex raises ValueError in some cases when the
                       # indices do not intersect (see the docstring of
                       # as_subindex())
                       continue

                   if not index.isempty(self):
                       # Yield those c for which idx.as_subindex(c) is nonempty
                       yield c

        except it is more efficient.

        >>> from ndindex import ChunkSize, Tuple
        >>> idx = Tuple(slice(5, 15), 0)
        >>> shape = (20, 20)
        >>> chunk_size = ChunkSize((10, 10))
        >>> for c in chunk_size.as_subchunks(idx, shape):
        ...     print(c)
        ...     print('    ', idx.as_subindex(c))
        Tuple(slice(0, 10, 1), slice(0, 10, 1))
            Tuple(slice(5, 10, 1), 0)
        Tuple(slice(10, 20, 1), slice(0, 10, 1))
            Tuple(slice(0, 5, 1), 0)

        See Also
        ========

        ndindex.NDIndex.as_subindex
        num_subchunks

        """
        shape = asshape(shape)
        if len(shape) != len(self):
            raise ValueError("chunks dimensions must equal the array dimensions")

        if 0 in shape:
            return
        idx = ndindex(idx).expand(shape)

        # The slow naive fallback is kept here for testing purposes and to support
        # indices that aren't supported in the fast way yet below.
        def _fallback():
            if _force_slow is False: # pragma: no cover
                raise RuntimeError("as_subchunks() attempted fallback with _force_slow=False")
            for c in self.indices(shape):
                try:
                    index = idx.as_subindex(c)
                except ValueError:
                    continue

                if not index.isempty(self):
                    yield c
            return

        if _force_slow:
            yield from _fallback()
            return

        if idx.isempty(shape):
            return

        if self == ():
            yield Tuple()
            return

        iters = []
        idx_args = iter(idx.args)
        self_ = iter(self)
        while True:
            try:
                i = next(idx_args)
                if isinstance(i, Newaxis) or i == True:
                    continue
                n = next(self_)
            except StopIteration:
                break
            if isinstance(i, Integer):
                iters.append([i.raw//n])
            elif isinstance(i, IntegerArray):
                # All arrays will be together after calling expand() (Tuple does not support arrays
                # separated by non-integer indices). Collect them all together
                # at once.
                arrs = []
                while True:
                    try:
                        if isinstance(i, IntegerArray):
                            arrs.append(i.raw.flatten()//n)
                        else:
                            idx_args = chain([i], idx_args)
                            self_ = chain([n], self_)
                            break
                        i = next(idx_args)
                        n = next(self_)
                    except StopIteration:
                        break

                import numpy as np
                a = np.unique(np.stack(arrs), axis=-1)
                def _array_iter(a):
                    for i in range(a.shape[-1]):
                        yield tuple(a[..., i].flat)
                iters.append(_array_iter(a))
            elif isinstance(i, Slice) and i.step > 0:
                def _slice_iter(s, n):
                    a, N, m = s.args
                    if m > n:
                        yield from ((a + k*m)//n for k in range(ceiling(N-a, m)))
                    else:
                        yield from range(a//n, ceiling(N, n))
                iters.append(_slice_iter(i, n))
            else:
                # Fallback to the naive algorithm. This should currently only
                # happen in cases where the naive as_subindex algorithm will
                # raise NotImplementedError.
                yield from _fallback()
                return # pragma: no cover

        def _flatten(l):
            for element in l:
                if isinstance(element, tuple):
                    yield from element
                else:
                    yield element

        def _indices(iters):
            for _p in product(*iters):
                p = _flatten(_p)
                # p = (0, 0, 0), (0, 0, 1), ...
                yield Tuple(*[Slice(chunk_size*i, min(chunk_size*(i + 1), n), 1)
                          for n, chunk_size, i in zip(shape, self, p)])

        for c in _indices(iters):
            # Empty indices should be impossible by the construction of the
            # iterators above.
            yield c

    def num_subchunks(self, idx, shape):
        """
        Give the number of chunks indexed by `idx` on an array of shape
        `shape`.

        This is equivalent to `len(list(self.as_subindex(idx, shape)))`, but
        more efficient.

        >>> from ndindex import ChunkSize, Tuple
        >>> idx = Tuple(slice(5, 15), 0)
        >>> shape = (20, 20)
        >>> chunk_size = ChunkSize((10, 10))
        >>> chunk_size.num_subchunks(idx, shape)
        2

        See Also
        ========

        ndindex.NDIndex.as_subindex
        num_subchunks

        """
        shape = asshape(shape)
        if len(shape) != len(self):
            raise ValueError("chunks dimensions must equal the array dimensions")

        idx = ndindex(idx).expand(shape)

        if 0 in shape:
            return 0

        if idx.isempty(shape):
            return 0

        if self == ():
            return 1

        idx_args = iter(idx.args)
        self_ = iter(self)
        res = 1
        while True:
            try:
                i = next(idx_args)
                if isinstance(i, Newaxis) or i == True:
                    continue
                n = next(self_)
            except StopIteration:
                break
            if isinstance(i, Integer):
                continue
            elif isinstance(i, IntegerArray):
                arrs = []
                # see as_subchunks
                while True:
                    try:
                        if isinstance(i, IntegerArray):
                            arrs.append(i.raw.flatten()//n)
                        else:
                            idx_args = chain([i], idx_args)
                            self_ = chain([n], self_)
                            break
                        i = next(idx_args)
                        n = next(self_)
                    except StopIteration:
                        break

                import numpy as np
                res *= np.unique(np.stack(arrs), axis=-1).shape[-1]
            elif isinstance(i, Slice):
                if i.step < 0:
                    raise NotImplementedError("num_subchunks() is not implemented for slices with negative step")
                a, N, m = i.args
                if m > n:
                    res *= ceiling(N-a, m)
                else:
                    res *= ceiling(N, n) - a//n
            else:
                raise NotImplementedError(f"num_subchunks() is not implemented for {type(i).__name__}")

        return res

    def containing_block(self, idx, shape):
        """
        Compute the index for the smallest contiguous block of chunks that
        contains `idx` on an array of shape `shape`.

        A block is a subset of an array that is contiguous in all dimensions
        and is aligned along the chunk size. A block index is always of the
        form `(Slice(k1, m1), Slice(k2, m2), …, Slice(kn, mn))` where `n` is
        the number of dimensions in the chunk size, and the `ki` and `mi` are
        multiples of the corresponding chunk dimension (the `mi` may be
        truncated to the shape).

        For example, given a chunk size of `(10, 15)`, an example block might
        be `(Slice(0, 20), Slice(30, 45))`. Such a block would be the smallest
        block that contains the index `(Slice(0, 12), 40)`, for example.

        >>> from ndindex import ChunkSize
        >>> chunk_size = ChunkSize((10, 15))
        >>> idx = (slice(0, 12), 40)
        >>> shape = (100, 100)
        >>> block = chunk_size.containing_block(idx, shape)
        >>> block
        Tuple(slice(0, 20, 1), slice(30, 45, 1))

        The method :meth:`as_subchunks` can be used on the block to determine
        which chunks are contained in it, and :meth:`num_subchunks` to
        determine how many:

        >>> chunk_size.num_subchunks(block, shape)
        2
        >>> for c in chunk_size.as_subchunks(block, shape):
        ...     print(c)
        Tuple(slice(0, 10, 1), slice(30, 45, 1))
        Tuple(slice(10, 20, 1), slice(30, 45, 1))

        In this example, `chunk_size.as_subchunk(block, shape)` and
        `chunk_size.as_subchunks(idx, shape)` are the same, but in general, a
        block may overlap with more chunks than the original index because the
        block is contiguous.

        """
        shape = asshape(shape)
        if len(shape) != len(self):
            raise ValueError("chunks dimensions must equal the array dimensions")
        idx = ndindex(idx).expand(shape)

        idx_args = iter(idx.args)
        self_ = iter(self)
        res = []

        if False in idx.args:
            return Tuple(*[slice(0, 0, 1) for i in range(len(shape))])

        while True:
            try:
                i = next(idx_args)
                if isinstance(i, Newaxis) or i == True:
                    continue
                n = next(self_)
            except StopIteration:
                break
            if isinstance(i, Integer):
                chunk_n = i.raw//n
                res.append(Slice(chunk_n*n, (chunk_n + 1)*n))
            elif isinstance(i, IntegerArray):
                import numpy as np

                if i.size == 0:
                    res.append(Slice(0, 0))
                    continue
                m = np.min(i.array)
                M = np.max(i.array)
                res.append(Slice(m//n*n, (M//n + 1)*n))
            elif isinstance(i, Slice):
                if i.step < 0:
                    raise NotImplementedError("containing_block() is not implemented for slices with negative step")
                res.append(Slice(i.start - (i.start % n), ceiling(i.stop, n)*n))
            else:
                raise NotImplementedError(f"containing_block() is not implemented for {type(i).__name__}")

        return Tuple(*res).expand(shape)
